"""UniFi Protect Data."""

from __future__ import annotations

import asyncio
import logging
import zoneinfo
from collections.abc import Callable
from datetime import datetime, timedelta, tzinfo
from functools import cache
from ipaddress import IPv4Address, IPv6Address
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Literal
from uuid import UUID

import aiofiles
import orjson
from aiofiles import os as aos
from convertertools import pop_dict_set_if_none, pop_dict_tuple
from pydantic.v1.fields import PrivateAttr

from ..exceptions import BadRequest, NotAuthorized
from ..utils import RELEASE_CACHE, convert_to_datetime
from .base import (
    ProtectBaseObject,
    ProtectDeviceModel,
    ProtectModelWithId,
)
from .devices import (
    Camera,
    CameraZone,
    Light,
    OSDSettings,
    RecordingSettings,
    Sensor,
    SmartDetectSettings,
)
from .types import (
    AnalyticsOption,
    DoorbellMessageType,
    DoorbellText,
    EventCategories,
    EventType,
    FirmwareReleaseChannel,
    IteratorCallback,
    ModelType,
    MountType,
    PercentFloat,
    PercentInt,
    PermissionNode,
    ProgressCallback,
    RecordingMode,
    RecordingType,
    ResolutionStorageType,
    SensorStatusType,
    SensorType,
    SmartDetectObjectType,
    StorageType,
    Version,
)
from .user import User, UserLocation

if TYPE_CHECKING:
    from pydantic.v1.typing import SetStr


_LOGGER = logging.getLogger(__name__)
MAX_SUPPORTED_CAMERAS = 256
MAX_EVENT_HISTORY_IN_STATE_MACHINE = MAX_SUPPORTED_CAMERAS * 2
DELETE_KEYS_THUMB = {"color", "vehicleType"}
DELETE_KEYS_EVENT = {"deletedAt", "category", "subCategory"}


class NVRLocation(UserLocation):
    is_geofencing_enabled: bool
    radius: int
    model: ModelType | None = None


class SmartDetectItem(ProtectBaseObject):
    id: str
    timestamp: datetime
    level: PercentInt
    coord: tuple[int, int, int, int]
    object_type: SmartDetectObjectType
    zone_ids: list[int]
    duration: timedelta

    @classmethod
    @cache
    def _get_unifi_remaps(cls) -> dict[str, str]:
        return {
            **super()._get_unifi_remaps(),
            "zones": "zoneIds",
        }

    @classmethod
    @cache
    def unifi_dict_conversions(cls) -> dict[str, object | Callable[[Any], Any]]:
        return {
            "duration": lambda x: timedelta(milliseconds=x),
        } | super().unifi_dict_conversions()


class SmartDetectTrack(ProtectBaseObject):
    id: str
    payload: list[SmartDetectItem]
    camera_id: str
    event_id: str

    @classmethod
    @cache
    def _get_unifi_remaps(cls) -> dict[str, str]:
        return {
            **super()._get_unifi_remaps(),
            "camera": "cameraId",
            "event": "eventId",
        }

    @property
    def camera(self) -> Camera:
        return self._api.bootstrap.cameras[self.camera_id]

    @property
    def event(self) -> Event | None:
        return self._api.bootstrap.events.get(self.event_id)


class LicensePlateMetadata(ProtectBaseObject):
    name: str
    confidence_level: int


class EventThumbnailAttribute(ProtectBaseObject):
    confidence: int
    val: str


class NfcMetadata(ProtectBaseObject):
    nfc_id: str
    user_id: str

    @classmethod
    @cache
    def _get_unifi_remaps(cls) -> dict[str, str]:
        return {
            **super()._get_unifi_remaps(),
            "nfcId": "nfc_id",
            "userId": "user_id",
        }


class FingerprintMetadata(ProtectBaseObject):
    ulp_id: str | None = None

    @classmethod
    @cache
    def _get_unifi_remaps(cls) -> dict[str, str]:
        return {
            **super()._get_unifi_remaps(),
            "ulpId": "ulp_id",
        }


class EventThumbnailAttributes(ProtectBaseObject):
    color: EventThumbnailAttribute | None = None
    vehicle_type: EventThumbnailAttribute | None = None

    def unifi_dict(
        self,
        data: dict[str, Any] | None = None,
        exclude: set[str] | None = None,
    ) -> dict[str, Any]:
        data = super().unifi_dict(data=data, exclude=exclude)
        pop_dict_set_if_none(data, DELETE_KEYS_THUMB)
        return data


class EventDetectedThumbnail(ProtectBaseObject):
    clock_best_wall: datetime | None = None
    type: str
    cropped_id: str
    attributes: EventThumbnailAttributes | None = None
    name: str | None

    @classmethod
    @cache
    def unifi_dict_conversions(cls) -> dict[str, object | Callable[[Any], Any]]:
        return {"clockBestWall": convert_to_datetime} | super().unifi_dict_conversions()

    def unifi_dict(
        self,
        data: dict[str, Any] | None = None,
        exclude: set[str] | None = None,
    ) -> dict[str, Any]:
        data = super().unifi_dict(data=data, exclude=exclude)
        pop_dict_set_if_none(data, {"name"})
        return data


class EventMetadata(ProtectBaseObject):
    client_platform: str | None
    reason: str | None
    app_update: str | None
    light_id: str | None
    light_name: str | None
    type: str | None
    sensor_id: str | None
    sensor_name: str | None
    sensor_type: SensorType | None
    doorlock_id: str | None
    doorlock_name: str | None
    from_value: str | None
    to_value: str | None
    mount_type: MountType | None
    status: SensorStatusType | None
    alarm_type: str | None
    device_id: str | None
    mac: str | None
    # require 2.7.5+
    license_plate: LicensePlateMetadata | None = None
    # requires 2.11.13+
    detected_thumbnails: list[EventDetectedThumbnail] | None = None
    # requires 5.1.34+
    nfc: NfcMetadata | None = None
    fingerprint: FingerprintMetadata | None = None

    _collapse_keys: ClassVar[SetStr] = {
        "lightId",
        "lightName",
        "type",
        "sensorId",
        "sensorName",
        "sensorType",
        "doorlockId",
        "doorlockName",
        "mountType",
        "status",
        "alarmType",
        "deviceId",
        "mac",
    }

    @classmethod
    @cache
    def _get_unifi_remaps(cls) -> dict[str, str]:
        return {
            **super()._get_unifi_remaps(),
            "from": "fromValue",
            "to": "toValue",
        }

    @classmethod
    def unifi_dict_to_dict(cls, data: dict[str, Any]) -> dict[str, Any]:
        for key in cls._collapse_keys.intersection(data):
            if isinstance(data[key], dict):
                data[key] = data[key]["text"]

        return super().unifi_dict_to_dict(data)

    def unifi_dict(
        self,
        data: dict[str, Any] | None = None,
        exclude: set[str] | None = None,
    ) -> dict[str, Any]:
        data = super().unifi_dict(data=data, exclude=exclude)

        # all metadata keys optionally appear
        for key, value in list(data.items()):
            if value is None:
                del data[key]

        for key in self._collapse_keys.intersection(data):
            # AI Theta/Hotplug exception
            if key != "type" or data[key] not in {"audio", "video", "extender"}:
                data[key] = {"text": data[key]}

        return data


class Event(ProtectModelWithId):
    type: EventType
    start: datetime
    end: datetime | None
    score: int
    heatmap_id: str | None
    camera_id: str | None
    smart_detect_types: list[SmartDetectObjectType]
    smart_detect_event_ids: list[str]
    thumbnail_id: str | None
    user_id: str | None
    timestamp: datetime | None
    metadata: EventMetadata | None
    # requires 2.7.5+
    deleted_at: datetime | None = None
    deletion_type: Literal["manual", "automatic"] | None = None
    # only appears if `get_events` is called with category
    category: EventCategories | None = None
    sub_category: str | None = None

    # TODO:
    # partition
    # description

    _smart_detect_events: list[Event] | None = PrivateAttr(None)
    _smart_detect_track: SmartDetectTrack | None = PrivateAttr(None)
    _smart_detect_zones: dict[int, CameraZone] | None = PrivateAttr(None)

    @classmethod
    @cache
    def _get_unifi_remaps(cls) -> dict[str, str]:
        return {
            **super()._get_unifi_remaps(),
            "camera": "cameraId",
            "heatmap": "heatmapId",
            "user": "userId",
            "thumbnail": "thumbnailId",
            "smartDetectEvents": "smartDetectEventIds",
        }

    @classmethod
    @cache
    def unifi_dict_conversions(cls) -> dict[str, object | Callable[[Any], Any]]:
        return {
            key: convert_to_datetime
            for key in ("start", "end", "timestamp", "deletedAt")
        } | super().unifi_dict_conversions()

    def unifi_dict(
        self,
        data: dict[str, Any] | None = None,
        exclude: set[str] | None = None,
    ) -> dict[str, Any]:
        data = super().unifi_dict(data=data, exclude=exclude)
        pop_dict_set_if_none(data, DELETE_KEYS_EVENT)
        return data

    @property
    def camera(self) -> Camera | None:
        if self.camera_id is None:
            return None

        return self._api.bootstrap.cameras.get(self.camera_id)

    @property
    def light(self) -> Light | None:
        if self.metadata is None or self.metadata.light_id is None:
            return None

        return self._api.bootstrap.lights.get(self.metadata.light_id)

    @property
    def sensor(self) -> Sensor | None:
        if self.metadata is None or self.metadata.sensor_id is None:
            return None

        return self._api.bootstrap.sensors.get(self.metadata.sensor_id)

    @property
    def user(self) -> User | None:
        if self.user_id is None:
            return None

        return self._api.bootstrap.users.get(self.user_id)

    @property
    def smart_detect_events(self) -> list[Event]:
        if self._smart_detect_events is not None:
            return self._smart_detect_events

        self._smart_detect_events = [
            self._api.bootstrap.events[g]
            for g in self.smart_detect_event_ids
            if g in self._api.bootstrap.events
        ]
        return self._smart_detect_events

    async def get_thumbnail(
        self,
        width: int | None = None,
        height: int | None = None,
    ) -> bytes | None:
        """Gets thumbnail for event"""
        if self.thumbnail_id is None:
            return None
        if not self._api.bootstrap.auth_user.can(
            ModelType.CAMERA,
            PermissionNode.READ_MEDIA,
            self.camera,
        ):
            raise NotAuthorized(
                f"Do not have permission to read media for camera: {self.id}",
            )
        return await self._api.get_event_thumbnail(self.thumbnail_id, width, height)

    async def get_animated_thumbnail(
        self,
        width: int | None = None,
        height: int | None = None,
        *,
        speedup: int = 10,
    ) -> bytes | None:
        """Gets animated thumbnail for event"""
        if self.thumbnail_id is None:
            return None
        if not self._api.bootstrap.auth_user.can(
            ModelType.CAMERA,
            PermissionNode.READ_MEDIA,
            self.camera,
        ):
            raise NotAuthorized(
                f"Do not have permission to read media for camera: {self.id}",
            )
        return await self._api.get_event_animated_thumbnail(
            self.thumbnail_id,
            width,
            height,
            speedup=speedup,
        )

    async def get_heatmap(self) -> bytes | None:
        """Gets heatmap for event"""
        if self.heatmap_id is None:
            return None
        if not self._api.bootstrap.auth_user.can(
            ModelType.CAMERA,
            PermissionNode.READ_MEDIA,
            self.camera,
        ):
            raise NotAuthorized(
                f"Do not have permission to read media for camera: {self.id}",
            )
        return await self._api.get_event_heatmap(self.heatmap_id)

    async def get_video(
        self,
        channel_index: int = 0,
        output_file: Path | None = None,
        iterator_callback: IteratorCallback | None = None,
        progress_callback: ProgressCallback | None = None,
        chunk_size: int = 65536,
    ) -> bytes | None:
        """
        Get the MP4 video clip for this given event

        Args:
        ----
            channel_index: index of `CameraChannel` on the camera to use to retrieve video from

        Will raise an exception if event does not have a camera, end time or the channel index is wrong.

        """
        if self.camera is None:
            raise BadRequest("Event does not have a camera")
        if self.end is None:
            raise BadRequest("Event is ongoing")

        if not self._api.bootstrap.auth_user.can(
            ModelType.CAMERA,
            PermissionNode.READ_MEDIA,
            self.camera,
        ):
            raise NotAuthorized(
                f"Do not have permission to read media for camera: {self.id}",
            )
        return await self._api.get_camera_video(
            self.camera.id,
            self.start,
            self.end,
            channel_index,
            output_file=output_file,
            iterator_callback=iterator_callback,
            progress_callback=progress_callback,
            chunk_size=chunk_size,
        )

    async def get_smart_detect_track(self) -> SmartDetectTrack:
        """
        Gets smart detect track for given smart detect event.

        If event is not a smart detect event, it will raise a `BadRequest`
        """
        if self.type not in {EventType.SMART_DETECT, EventType.SMART_DETECT_LINE}:
            raise BadRequest("Not a smart detect event")

        if self._smart_detect_track is None:
            self._smart_detect_track = await self._api.get_event_smart_detect_track(
                self.id,
            )

        return self._smart_detect_track

    async def get_smart_detect_zones(self) -> dict[int, CameraZone]:
        """Gets the triggering zones for the smart detection"""
        if self.camera is None:
            raise BadRequest("No camera on event")

        if self._smart_detect_zones is None:
            smart_track = await self.get_smart_detect_track()

            ids: set[int] = set()
            for item in smart_track.payload:
                ids |= set(item.zone_ids)

            self._smart_detect_zones = {
                z.id: z for z in self.camera.smart_detect_zones if z.id in ids
            }

        return self._smart_detect_zones


class PortConfig(ProtectBaseObject):
    ump: int
    http: int
    https: int
    rtsp: int
    rtsps: int
    rtmp: int
    devices_wss: int
    camera_https: int
    live_ws: int
    live_wss: int
    tcp_streams: int
    playback: int
    ems_cli: int
    ems_live_flv: int
    camera_events: int
    tcp_bridge: int
    ucore: int
    discovery_client: int
    piongw: int | None = None
    ems_json_cli: int | None = None
    stacking: int | None = None
    # 3.0.22+
    ai_feature_console: int | None = None

    @classmethod
    @cache
    def _get_unifi_remaps(cls) -> dict[str, str]:
        return {
            **super()._get_unifi_remaps(),
            "emsCLI": "emsCli",
            "emsLiveFLV": "emsLiveFlv",
            "emsJsonCLI": "emsJsonCli",
        }


class CPUInfo(ProtectBaseObject):
    average_load: float
    temperature: float


class MemoryInfo(ProtectBaseObject):
    available: int | None
    free: int | None
    total: int | None


class StorageDevice(ProtectBaseObject):
    model: str
    size: int
    healthy: bool | str


class StorageInfo(ProtectBaseObject):
    available: int
    is_recycling: bool
    size: int
    type: StorageType
    used: int
    devices: list[StorageDevice]
    # requires 2.8.14+
    capability: str | None = None

    @classmethod
    def unifi_dict_to_dict(cls, data: dict[str, Any]) -> dict[str, Any]:
        if "type" in data:
            storage_type = data.pop("type")
            try:
                data["type"] = StorageType(storage_type)
            except ValueError:
                _LOGGER.warning("Unknown storage type: %s", storage_type)
                data["type"] = StorageType.UNKNOWN

        return super().unifi_dict_to_dict(data)


class StorageSpace(ProtectBaseObject):
    total: int
    used: int
    available: int


class TMPFSInfo(ProtectBaseObject):
    available: int
    total: int
    used: int
    path: Path


class UOSDisk(ProtectBaseObject):
    slot: int
    state: str

    type: Literal["SSD", "HDD"] | None = None
    model: str | None = None
    serial: str | None = None
    firmware: str | None = None
    rpm: int | None = None
    ata: str | None = None
    sata: str | None = None
    action: str | None = None
    healthy: str | None = None
    reason: list[Any] | None = None
    temperature: int | None = None
    power_on_hours: int | None = None
    life_span: PercentFloat | None = None
    bad_sector: int | None = None
    threshold: int | None = None
    progress: PercentFloat | None = None
    estimate: timedelta | None = None
    # 2.10.10+
    size: int | None = None

    @classmethod
    @cache
    def _get_unifi_remaps(cls) -> dict[str, str]:
        return {
            **super()._get_unifi_remaps(),
            "poweronhrs": "powerOnHours",
            "life_span": "lifeSpan",
            "bad_sector": "badSector",
        }

    @classmethod
    @cache
    def unifi_dict_conversions(cls) -> dict[str, object | Callable[[Any], Any]]:
        return {
            "estimate": lambda x: timedelta(seconds=x)
        } | super().unifi_dict_conversions()

    def unifi_dict(
        self,
        data: dict[str, Any] | None = None,
        exclude: set[str] | None = None,
    ) -> dict[str, Any]:
        data = super().unifi_dict(data=data, exclude=exclude)

        # estimate is actually in seconds, not milliseconds
        if "estimate" in data and data["estimate"] is not None:
            data["estimate"] /= 1000

        if "state" in data and data["state"] == "nodisk":
            pop_dict_tuple(
                data,
                (
                    "action",
                    "ata",
                    "bad_sector",
                    "estimate",
                    "firmware",
                    "healthy",
                    "life_span",
                    "model",
                    "poweronhrs",
                    "progress",
                    "reason",
                    "rpm",
                    "sata",
                    "serial",
                    "tempature",
                    "temperature",
                    "threshold",
                    "type",
                ),
            )
        return data

    @property
    def has_disk(self) -> bool:
        return self.state != "nodisk"

    @property
    def is_healthy(self) -> bool:
        return self.state in {
            "initializing",
            "expanding",
            "spare",
            "normal",
        }


class UOSSpace(ProtectBaseObject):
    device: str
    total_bytes: int
    used_bytes: int
    action: str
    progress: PercentFloat | None = None
    estimate: timedelta | None = None
    # requires 2.8.14+
    health: str | None = None
    # requires 2.8.22+
    space_type: str | None = None

    # TODO:
    # reasons

    @classmethod
    @cache
    def _get_unifi_remaps(cls) -> dict[str, str]:
        return {
            **super()._get_unifi_remaps(),
            "total_bytes": "totalBytes",
            "used_bytes": "usedBytes",
            "space_type": "spaceType",
        }

    @classmethod
    @cache
    def unifi_dict_conversions(cls) -> dict[str, object | Callable[[Any], Any]]:
        return {
            "estimate": lambda x: timedelta(seconds=x)
        } | super().unifi_dict_conversions()

    def unifi_dict(
        self,
        data: dict[str, Any] | None = None,
        exclude: set[str] | None = None,
    ) -> dict[str, Any]:
        data = super().unifi_dict(data=data, exclude=exclude)

        # estimate is actually in seconds, not milliseconds
        if "estimate" in data and data["estimate"] is not None:
            data["estimate"] /= 1000

        return data


class UOSStorage(ProtectBaseObject):
    disks: list[UOSDisk]
    space: list[UOSSpace]

    # TODO:
    # sdcards


class SystemInfo(ProtectBaseObject):
    cpu: CPUInfo
    memory: MemoryInfo
    storage: StorageInfo
    tmpfs: TMPFSInfo
    ustorage: UOSStorage | None = None

    def unifi_dict(
        self,
        data: dict[str, Any] | None = None,
        exclude: set[str] | None = None,
    ) -> dict[str, Any]:
        data = super().unifi_dict(data=data, exclude=exclude)
        pop_dict_set_if_none(data, {"ustorage"})
        return data


class DoorbellMessage(ProtectBaseObject):
    type: DoorbellMessageType
    text: DoorbellText


class DoorbellSettings(ProtectBaseObject):
    default_message_text: DoorbellText
    default_message_reset_timeout: timedelta
    all_messages: list[DoorbellMessage]
    custom_messages: list[DoorbellText]

    @classmethod
    @cache
    def _get_unifi_remaps(cls) -> dict[str, str]:
        return {
            **super()._get_unifi_remaps(),
            "defaultMessageResetTimeoutMs": "defaultMessageResetTimeout",
        }

    @classmethod
    @cache
    def unifi_dict_conversions(cls) -> dict[str, object | Callable[[Any], Any]]:
        return {
            # defaultMessageResetTimeoutMs is remapped to defaultMessageResetTimeout
            "defaultMessageResetTimeoutMs": lambda x: timedelta(milliseconds=x),
        } | super().unifi_dict_conversions()


class RecordingTypeDistribution(ProtectBaseObject):
    recording_type: RecordingType
    size: int
    percentage: float


class ResolutionDistribution(ProtectBaseObject):
    resolution: ResolutionStorageType
    size: int
    percentage: float


class StorageDistribution(ProtectBaseObject):
    recording_type_distributions: list[RecordingTypeDistribution]
    resolution_distributions: list[ResolutionDistribution]

    _recording_type_dict: dict[RecordingType, RecordingTypeDistribution] | None = (
        PrivateAttr(None)
    )
    _resolution_dict: dict[ResolutionStorageType, ResolutionDistribution] | None = (
        PrivateAttr(None)
    )

    def _get_recording_type_dict(
        self,
    ) -> dict[RecordingType, RecordingTypeDistribution]:
        if self._recording_type_dict is None:
            self._recording_type_dict = {}
            for recording_type in self.recording_type_distributions:
                self._recording_type_dict[recording_type.recording_type] = (
                    recording_type
                )

        return self._recording_type_dict

    def _get_resolution_dict(
        self,
    ) -> dict[ResolutionStorageType, ResolutionDistribution]:
        if self._resolution_dict is None:
            self._resolution_dict = {}
            for resolution in self.resolution_distributions:
                self._resolution_dict[resolution.resolution] = resolution

        return self._resolution_dict

    @property
    def timelapse_recordings(self) -> RecordingTypeDistribution | None:
        return self._get_recording_type_dict().get(RecordingType.TIMELAPSE)

    @property
    def continuous_recordings(self) -> RecordingTypeDistribution | None:
        return self._get_recording_type_dict().get(RecordingType.CONTINUOUS)

    @property
    def detections_recordings(self) -> RecordingTypeDistribution | None:
        return self._get_recording_type_dict().get(RecordingType.DETECTIONS)

    @property
    def uhd_usage(self) -> ResolutionDistribution | None:
        return self._get_resolution_dict().get(ResolutionStorageType.UHD)

    @property
    def hd_usage(self) -> ResolutionDistribution | None:
        return self._get_resolution_dict().get(ResolutionStorageType.HD)

    @property
    def free(self) -> ResolutionDistribution | None:
        return self._get_resolution_dict().get(ResolutionStorageType.FREE)

    def update_from_dict(self, data: dict[str, Any]) -> StorageDistribution:
        # reset internal look ups when data changes
        self._recording_type_dict = None
        self._resolution_dict = None

        return super().update_from_dict(data)


class StorageStats(ProtectBaseObject):
    utilization: float
    capacity: timedelta | None
    remaining_capacity: timedelta | None
    recording_space: StorageSpace
    storage_distribution: StorageDistribution

    @classmethod
    @cache
    def unifi_dict_conversions(cls) -> dict[str, object | Callable[[Any], Any]]:
        return {
            "capacity": lambda x: timedelta(milliseconds=x),
            "remainingCapacity": lambda x: timedelta(milliseconds=x),
        } | super().unifi_dict_conversions()


class NVRFeatureFlags(ProtectBaseObject):
    beta: bool
    dev: bool
    notifications_v2: bool
    homekit_paired: bool | None = None
    ulp_role_management: bool | None = None
    # 2.9.20+
    detection_labels: bool | None = None
    has_two_way_audio_media_streams: bool | None = None


class NVRSmartDetection(ProtectBaseObject):
    enable: bool
    face_recognition: bool
    license_plate_recognition: bool


class GlobalRecordingSettings(ProtectBaseObject):
    osd_settings: OSDSettings
    recording_settings: RecordingSettings
    smart_detect_settings: SmartDetectSettings

    # TODO:
    # recordingSchedulesV2


class NVR(ProtectDeviceModel):
    can_auto_update: bool
    is_stats_gathering_enabled: bool
    timezone: tzinfo
    version: Version
    ucore_version: str
    hardware_platform: str
    ports: PortConfig
    last_update_at: datetime | None
    is_station: bool
    enable_automatic_backups: bool
    enable_stats_reporting: bool
    release_channel: FirmwareReleaseChannel
    hosts: list[IPv4Address | IPv6Address | str]
    enable_bridge_auto_adoption: bool
    hardware_id: UUID
    host_type: int
    host_shortname: str
    is_hardware: bool
    is_wireless_uplink_enabled: bool | None
    time_format: Literal["12h", "24h"]
    temperature_unit: Literal["C", "F"]
    recording_retention_duration: timedelta | None
    enable_crash_reporting: bool
    disable_audio: bool
    analytics_data: AnalyticsOption
    anonymous_device_id: UUID | None
    camera_utilization: int
    is_recycling: bool
    disable_auto_link: bool
    skip_firmware_update: bool
    location_settings: NVRLocation
    feature_flags: NVRFeatureFlags
    system_info: SystemInfo
    doorbell_settings: DoorbellSettings
    storage_stats: StorageStats
    is_away: bool
    is_setup: bool
    network: str
    max_camera_capacity: dict[Literal["4K", "2K", "HD"], int]
    market_name: str | None = None
    stream_sharing_available: bool | None = None
    is_db_available: bool | None = None
    is_insights_enabled: bool | None = None
    is_recording_disabled: bool | None = None
    is_recording_motion_only: bool | None = None
    ui_version: str | None = None
    sso_channel: FirmwareReleaseChannel | None = None
    is_stacked: bool | None = None
    is_primary: bool | None = None
    last_drive_slow_event: datetime | None = None
    is_u_core_setup: bool | None = None
    vault_camera_ids: list[str] = []
    # requires 2.8.14+
    corruption_state: str | None = None
    country_code: str | None = None
    has_gateway: bool | None = None
    is_vault_registered: bool | None = None
    public_ip: IPv4Address | None = None
    ulp_version: str | None = None
    wan_ip: IPv4Address | IPv6Address | None = None
    # requires 2.9.20+
    hard_drive_state: str | None = None
    is_network_installed: bool | None = None
    is_protect_updatable: bool | None = None
    is_ucore_updatable: bool | None = None
    # requires 2.11.13+
    last_device_fw_updates_checked_at: datetime | None = None
    # requires 3.0.22+
    smart_detection: NVRSmartDetection | None = None
    is_ucore_stacked: bool | None = None
    global_camera_settings: GlobalRecordingSettings | None = None

    # TODO:
    # errorCode   read only
    # wifiSettings
    # smartDetectAgreement
    # dbRecoveryOptions
    # portStatus
    # cameraCapacity
    # deviceFirmwareSettings

    @classmethod
    @cache
    def _get_unifi_remaps(cls) -> dict[str, str]:
        return {
            **super()._get_unifi_remaps(),
            "recordingRetentionDurationMs": "recordingRetentionDuration",
            "vaultCameras": "vaultCameraIds",
            "lastDeviceFWUpdatesCheckedAt": "lastDeviceFwUpdatesCheckedAt",
            "isUCoreStacked": "isUcoreStacked",
        }

    @classmethod
    @cache
    def _get_read_only_fields(cls) -> set[str]:
        return super()._get_read_only_fields() | {
            "version",
            "uiVersion",
            "hardwarePlatform",
            "ports",
            "lastUpdateAt",
            "isStation",
            "hosts",
            "hostShortname",
            "isDbAvailable",
            "isRecordingDisabled",
            "isRecordingMotionOnly",
            "cameraUtilization",
            "storageStats",
            "isRecycling",
            "avgMotions",
            "streamSharingAvailable",
        }

    @classmethod
    @cache
    def unifi_dict_conversions(cls) -> dict[str, object | Callable[[Any], Any]]:
        return {
            "lastUpdateAt": convert_to_datetime,
            "lastDeviceFwUpdatesCheckedAt": convert_to_datetime,
            "timezone": zoneinfo.ZoneInfo,
            # recordingRetentionDurationMs is remapped to recordingRetentionDuration
            "recordingRetentionDurationMs": lambda x: timedelta(milliseconds=x),
        } | super().unifi_dict_conversions()

    async def _api_update(self, data: dict[str, Any]) -> None:
        return await self._api.update_nvr(data)

    @property
    def is_analytics_enabled(self) -> bool:
        return self.analytics_data is not AnalyticsOption.NONE

    @property
    def protect_url(self) -> str:
        return f"{self._api.base_url}/protect/devices/{self._api.bootstrap.nvr.id}"

    @property
    def display_name(self) -> str:
        return self.name or self.market_name or self.type

    @property
    def vault_cameras(self) -> list[Camera]:
        """Vault Cameras for NVR"""
        if not self.vault_camera_ids:
            return []
        return [self._api.bootstrap.cameras[c] for c in self.vault_camera_ids]

    @property
    def is_global_recording_enabled(self) -> bool:
        """
        Is recording footage/events from the camera enabled?

        If recording is not enabled, cameras will not produce any footage, thumbnails,
        motion/smart detection events.
        """
        return (
            (global_camera_settings := self.global_camera_settings) is not None
            and global_camera_settings.recording_settings.mode
            is not RecordingMode.NEVER
        )

    @property
    def is_smart_detections_enabled(self) -> bool:
        """If smart detected enabled globally."""
        return (
            smart_detection := self.smart_detection
        ) is not None and smart_detection.enable

    @property
    def is_license_plate_detections_enabled(self) -> bool:
        """If smart detected enabled globally."""
        return (
            (smart_detection := self.smart_detection) is not None
            and smart_detection.enable
            and smart_detection.license_plate_recognition
        )

    @property
    def is_face_detections_enabled(self) -> bool:
        """If smart detected enabled globally."""
        return (
            (smart_detection := self.smart_detection) is not None
            and smart_detection.enable
            and smart_detection.face_recognition
        )

    def update_all_messages(self) -> None:
        """Updates doorbell_settings.all_messages after adding/removing custom message"""
        messages = self.doorbell_settings.custom_messages
        self.doorbell_settings.all_messages = [
            DoorbellMessage(
                type=DoorbellMessageType.LEAVE_PACKAGE_AT_DOOR,
                text=DoorbellMessageType.LEAVE_PACKAGE_AT_DOOR.value.replace("_", " "),  # type: ignore[arg-type]
            ),
            DoorbellMessage(
                type=DoorbellMessageType.DO_NOT_DISTURB,
                text=DoorbellMessageType.DO_NOT_DISTURB.value.replace("_", " "),  # type: ignore[arg-type]
            ),
            *(
                DoorbellMessage(
                    type=DoorbellMessageType.CUSTOM_MESSAGE,
                    text=message,
                )
                for message in messages
            ),
        ]

    async def set_insights(self, enabled: bool) -> None:
        """Sets analytics collection for NVR"""

        def callback() -> None:
            self.is_insights_enabled = enabled

        await self.queue_update(callback)

    async def set_analytics(self, value: AnalyticsOption) -> None:
        """Sets analytics collection for NVR"""

        def callback() -> None:
            self.analytics_data = value

        await self.queue_update(callback)

    async def set_anonymous_analytics(self, enabled: bool) -> None:
        """Enables or disables anonymous analystics for NVR"""
        if enabled:
            await self.set_analytics(AnalyticsOption.ANONYMOUS)
        else:
            await self.set_analytics(AnalyticsOption.NONE)

    async def set_default_reset_timeout(self, timeout: timedelta) -> None:
        """Sets the default message reset timeout"""

        def callback() -> None:
            self.doorbell_settings.default_message_reset_timeout = timeout

        await self.queue_update(callback)

    async def set_default_doorbell_message(self, message: str) -> None:
        """Sets default doorbell message"""

        def callback() -> None:
            self.doorbell_settings.default_message_text = DoorbellText(message)

        await self.queue_update(callback)

    async def add_custom_doorbell_message(self, message: str) -> None:
        """Adds custom doorbell message"""
        if len(message) > 30:
            raise BadRequest("Message length over 30 characters")

        if message in self.doorbell_settings.custom_messages:
            raise BadRequest("Custom doorbell message already exists")

        await self._update_doorbell_messages(
            lambda: self.doorbell_settings.custom_messages.append(
                DoorbellText(message)
            ),
        )

    async def remove_custom_doorbell_message(self, message: str) -> None:
        """Removes custom doorbell message"""
        if message not in self.doorbell_settings.custom_messages:
            raise BadRequest("Custom doorbell message does not exists")

        await self._update_doorbell_messages(
            lambda: self.doorbell_settings.custom_messages.remove(
                DoorbellText(message)
            ),
        )

    async def _update_doorbell_messages(
        self, update_callback: Callable[[], None]
    ) -> None:
        """Updates doorbell messages and saves to Protect."""
        async with self._update_sync.lock:
            # yield to the event loop once we have the lock to ensure websocket updates are processed
            await asyncio.sleep(0)
            data_before_changes = self.dict_with_excludes()
            update_callback()
            await self.save_device(data_before_changes)
            self.update_all_messages()

    async def reboot(self) -> None:
        """Reboots the NVR"""
        await self._api.reboot_nvr()

    async def _read_cache_file(self, file_path: Path) -> set[Version] | None:
        versions: set[Version] | None = None
        try:
            _LOGGER.debug("Reading release cache file: %s", file_path)
            async with aiofiles.open(file_path, "rb") as cache_file:
                versions = {Version(v) for v in orjson.loads(await cache_file.read())}
        except FileNotFoundError:
            # ignore missing file
            pass
        except Exception:
            _LOGGER.warning("Failed to parse cache file: %s", file_path)

        return versions

    async def get_is_prerelease(self) -> bool:
        """Get if current version of Protect is a prerelease version."""
        # only EA versions have `-beta` in versions
        if self.version.is_prerelease:
            return True

        # 2.6.14 is an EA version that looks like a release version
        cache_file_path = self._api.cache_dir / "release_cache.json"
        versions = await self._read_cache_file(
            cache_file_path,
        ) or await self._read_cache_file(RELEASE_CACHE)
        if versions is None or self.version not in versions:
            versions = await self._api.get_release_versions()
            try:
                _LOGGER.debug("Fetching releases from APT repos...")
                tmp = self._api.cache_dir / "release_cache.tmp.json"
                await aos.makedirs(self._api.cache_dir, exist_ok=True)
                async with aiofiles.open(tmp, "wb") as cache_file:
                    await cache_file.write(orjson.dumps([str(v) for v in versions]))
                await aos.rename(tmp, cache_file_path)
            except Exception:
                _LOGGER.warning("Failed write cache file.")

        return self.version not in versions

    async def set_smart_detections(self, value: bool) -> None:
        """Set if smart detections are enabled."""

        def callback() -> None:
            if self.smart_detection is not None:
                self.smart_detection.enable = value

        await self.queue_update(callback)

    async def set_face_recognition(self, value: bool) -> None:
        """Set if face detections are enabled. Requires smart detections to be enabled."""
        if self.smart_detection is None or not self.smart_detection.enable:
            raise BadRequest("Smart detections are not enabled.")

        def callback() -> None:
            if self.smart_detection is not None:
                self.smart_detection.face_recognition = value

        await self.queue_update(callback)

    async def set_license_plate_recognition(self, value: bool) -> None:
        """Set if license plate detections are enabled. Requires smart detections to be enabled."""
        if self.smart_detection is None or not self.smart_detection.enable:
            raise BadRequest("Smart detections are not enabled.")

        def callback() -> None:
            if self.smart_detection is not None:
                self.smart_detection.license_plate_recognition = value

        await self.queue_update(callback)

    async def set_global_osd_name(self, enabled: bool) -> None:
        """Sets whether camera name is in the On Screen Display"""

        def callback() -> None:
            if self.global_camera_settings:
                self.global_camera_settings.osd_settings.is_name_enabled = enabled

        await self.queue_update(callback)

    async def set_global_osd_date(self, enabled: bool) -> None:
        """Sets whether current date is in the On Screen Display"""

        def callback() -> None:
            if self.global_camera_settings:
                self.global_camera_settings.osd_settings.is_date_enabled = enabled

        await self.queue_update(callback)

    async def set_global_osd_logo(self, enabled: bool) -> None:
        """Sets whether the UniFi logo is in the On Screen Display"""

        def callback() -> None:
            if self.global_camera_settings:
                self.global_camera_settings.osd_settings.is_logo_enabled = enabled

        await self.queue_update(callback)

    async def set_global_osd_bitrate(self, enabled: bool) -> None:
        """Sets whether camera bitrate is in the On Screen Display"""

        def callback() -> None:
            # mismatch between UI internal data structure debug = bitrate data
            if self.global_camera_settings:
                self.global_camera_settings.osd_settings.is_debug_enabled = enabled

        await self.queue_update(callback)

    async def set_global_motion_detection(self, enabled: bool) -> None:
        """Sets motion detection on camera"""

        def callback() -> None:
            if self.global_camera_settings:
                self.global_camera_settings.recording_settings.enable_motion_detection = enabled

        await self.queue_update(callback)

    async def set_global_recording_mode(self, mode: RecordingMode) -> None:
        """Sets recording mode on camera"""

        def callback() -> None:
            if self.global_camera_settings:
                self.global_camera_settings.recording_settings.mode = mode

        await self.queue_update(callback)

    # object smart detections

    def _is_smart_enabled(self, smart_type: SmartDetectObjectType) -> bool:
        return (
            self.is_global_recording_enabled
            and (global_camera_settings := self.global_camera_settings) is not None
            and smart_type in global_camera_settings.smart_detect_settings.object_types
        )

    @property
    def is_global_person_detection_on(self) -> bool:
        """
        Is Person Detection available and enabled (camera will produce person smart
        detection events)?
        """
        return self._is_smart_enabled(SmartDetectObjectType.PERSON)

    @property
    def is_global_person_tracking_enabled(self) -> bool:
        """Is person tracking enabled"""
        return (
            (global_camera_settings := self.global_camera_settings) is not None
            and (
                auto_tracking_object_types
                := global_camera_settings.smart_detect_settings.auto_tracking_object_types
            )
            is not None
            and SmartDetectObjectType.PERSON in auto_tracking_object_types
        )

    @property
    def is_global_vehicle_detection_on(self) -> bool:
        """
        Is Vehicle Detection available and enabled (camera will produce vehicle smart
        detection events)?
        """
        return self._is_smart_enabled(SmartDetectObjectType.VEHICLE)

    @property
    def is_global_license_plate_detection_on(self) -> bool:
        """
        Is License Plate Detection available and enabled (camera will produce face license
        plate detection events)?
        """
        return self._is_smart_enabled(SmartDetectObjectType.LICENSE_PLATE)

    @property
    def is_global_package_detection_on(self) -> bool:
        """
        Is Package Detection available and enabled (camera will produce package smart
        detection events)?
        """
        return self._is_smart_enabled(SmartDetectObjectType.PACKAGE)

    @property
    def is_global_animal_detection_on(self) -> bool:
        """
        Is Animal Detection available and enabled (camera will produce package smart
        detection events)?
        """
        return self._is_smart_enabled(SmartDetectObjectType.ANIMAL)

    def _is_audio_enabled(self, smart_type: SmartDetectObjectType) -> bool:
        return (
            (audio_type := smart_type.audio_type) is not None
            and self.is_global_recording_enabled
            and (global_camera_settings := self.global_camera_settings) is not None
            and (
                audio_types := global_camera_settings.smart_detect_settings.audio_types
            )
            is not None
            and audio_type in audio_types
        )

    @property
    def is_global_smoke_detection_on(self) -> bool:
        """
        Is Smoke Alarm Detection available and enabled (camera will produce smoke
        smart detection events)?
        """
        return self._is_audio_enabled(SmartDetectObjectType.SMOKE)

    @property
    def is_global_co_detection_on(self) -> bool:
        """
        Is CO Alarm Detection available and enabled (camera will produce smoke smart
        detection events)?
        """
        return self._is_audio_enabled(SmartDetectObjectType.CMONX)

    @property
    def is_global_siren_detection_on(self) -> bool:
        """
        Is Siren Detection available and enabled (camera will produce siren smart
        detection events)?
        """
        return self._is_audio_enabled(SmartDetectObjectType.SIREN)

    @property
    def is_global_baby_cry_detection_on(self) -> bool:
        """
        Is Baby Cry Detection available and enabled (camera will produce baby cry smart
        detection events)?
        """
        return self._is_audio_enabled(SmartDetectObjectType.BABY_CRY)

    @property
    def is_global_speaking_detection_on(self) -> bool:
        """
        Is Speaking Detection available and enabled (camera will produce speaking smart
        detection events)?
        """
        return self._is_audio_enabled(SmartDetectObjectType.SPEAK)

    @property
    def is_global_bark_detection_on(self) -> bool:
        """
        Is Bark Detection available and enabled (camera will produce barking smart
        detection events)?
        """
        return self._is_audio_enabled(SmartDetectObjectType.BARK)

    @property
    def is_global_car_alarm_detection_on(self) -> bool:
        """
        Is Car Alarm Detection available and enabled (camera will produce car alarm smart
        detection events)?
        """
        return self._is_audio_enabled(SmartDetectObjectType.BURGLAR)

    @property
    def is_global_car_horn_detection_on(self) -> bool:
        """
        Is Car Horn Detection available and enabled (camera will produce car horn smart
        detection events)?
        """
        return self._is_audio_enabled(SmartDetectObjectType.CAR_HORN)

    @property
    def is_global_glass_break_detection_on(self) -> bool:
        """
        Is Glass Break available and enabled (camera will produce glass break smart
        detection events)?
        """
        return self._is_audio_enabled(SmartDetectObjectType.GLASS_BREAK)


class LiveviewSlot(ProtectBaseObject):
    camera_ids: list[str]
    cycle_mode: str
    cycle_interval: int

    _cameras: list[Camera] | None = PrivateAttr(None)

    @classmethod
    @cache
    def _get_unifi_remaps(cls) -> dict[str, str]:
        return {**super()._get_unifi_remaps(), "cameras": "cameraIds"}

    @property
    def cameras(self) -> list[Camera]:
        if self._cameras is not None:
            return self._cameras

        # user may not have permission to see the cameras in the liveview
        self._cameras = [
            self._api.bootstrap.cameras[g]
            for g in self.camera_ids
            if g in self._api.bootstrap.cameras
        ]
        return self._cameras


class Liveview(ProtectModelWithId):
    name: str
    is_default: bool
    is_global: bool
    layout: int
    slots: list[LiveviewSlot]
    owner_id: str

    @classmethod
    @cache
    def _get_unifi_remaps(cls) -> dict[str, str]:
        return {**super()._get_unifi_remaps(), "owner": "ownerId"}

    @classmethod
    @cache
    def _get_read_only_fields(cls) -> set[str]:
        return super()._get_read_only_fields() | {"isDefault", "owner"}

    @property
    def owner(self) -> User | None:
        """
        Owner of liveview.

        Will be none if the user only has read only access and it was not made by their user.
        """
        return self._api.bootstrap.users.get(self.owner_id)

    @property
    def protect_url(self) -> str:
        return f"{self._api.base_url}/protect/liveview/{self.id}"



from __future__ import annotations

import asyncio
import logging
from asyncio.streams import StreamReader
from asyncio.subprocess import PIPE, Process, create_subprocess_exec
from pathlib import Path
from shlex import split
from typing import TYPE_CHECKING
from urllib.parse import urlparse

from aioshutil import which

from .exceptions import BadRequest, StreamError

if TYPE_CHECKING:
    from .data import Camera

_LOGGER = logging.getLogger(__name__)


class FfmpegCommand:
    ffmpeg_path: Path | None
    args: list[str]
    process: Process | None = None

    stdout: list[str] = []
    stderr: list[str] = []

    def __init__(self, cmd: str, ffmpeg_path: Path | None = None) -> None:
        self.args = split(cmd)

        if "ffmpeg" in self.args[0] and ffmpeg_path is None:
            self.ffmpeg_path = Path(self.args.pop(0))
        else:
            self.ffmpeg_path = ffmpeg_path

    @property
    def is_started(self) -> bool:
        return self.process is not None

    @property
    def is_running(self) -> bool:
        if self.process is None:
            return False

        return self.process.returncode is None

    @property
    def is_error(self) -> bool:
        if self.process is None:
            raise StreamError("ffmpeg has not started")

        if self.is_running:
            return False

        return self.process.returncode != 0

    async def start(self) -> None:
        if self.is_started:
            raise StreamError("ffmpeg command already started")

        if self.ffmpeg_path is None:
            system_ffmpeg = await which("ffmpeg")

            if system_ffmpeg is None:
                raise StreamError("Could not find ffmpeg")
            self.ffmpeg_path = Path(system_ffmpeg)

        if not self.ffmpeg_path.exists():
            raise StreamError("Could not find ffmpeg")

        _LOGGER.debug("ffmpeg: %s %s", self.ffmpeg_path, " ".join(self.args))
        self.process = await create_subprocess_exec(
            self.ffmpeg_path,
            *self.args,
            stdout=PIPE,
            stderr=PIPE,
        )

    async def stop(self) -> None:
        if self.process is None:
            raise StreamError("ffmpeg has not started")

        self.process.kill()
        await self.process.wait()

    async def _read_stream(self, stream: StreamReader | None, attr: str) -> None:
        if stream is None:
            return

        while True:
            line = await stream.readline()
            if line:
                getattr(self, attr).append(line.decode("utf8").rstrip())
            else:
                break

    async def run_until_complete(self) -> None:
        if not self.is_started:
            await self.start()

        if self.process is None:
            raise StreamError("Could not start stream")

        await asyncio.wait(
            [
                asyncio.create_task(self._read_stream(self.process.stdout, "stdout")),
                asyncio.create_task(self._read_stream(self.process.stderr, "stderr")),
            ],
        )
        await self.process.wait()


class TalkbackStream(FfmpegCommand):
    camera: Camera
    content_url: str

    def __init__(
        self,
        camera: Camera,
        content_url: str,
        ffmpeg_path: Path | None = None,
    ):
        if not camera.feature_flags.has_speaker:
            raise BadRequest("Camera does not have a speaker for talkback")

        content_url = self.clean_url(content_url)
        input_args = self.get_args_from_url(content_url)
        if len(input_args) > 0:
            input_args += " "

        # vn = no video
        # acodec = audio codec to encode output in (aac)
        # ac = number of output channels (1)
        # ar = output sampling rate (22050)
        # b:a = set bit rate of output audio
        cmd = (
            "-loglevel info -hide_banner "
            f'{input_args}-i "{content_url}" -vn '
            f"-acodec {camera.talkback_settings.type_fmt.value} -ac {camera.talkback_settings.channels} "
            f"-ar {camera.talkback_settings.sampling_rate} -b:a {camera.talkback_settings.sampling_rate} -map 0:a "
            f'-f adts "udp://{camera.host}:{camera.talkback_settings.bind_port}?bitrate={camera.talkback_settings.sampling_rate}"'
        )

        super().__init__(cmd, ffmpeg_path)

    @classmethod
    def clean_url(cls, content_url: str) -> str:
        parsed = urlparse(content_url)
        if parsed.scheme in {"file", ""}:
            path = Path(parsed.netloc + parsed.path)
            if not path.exists():
                raise BadRequest(f"File {path} does not exist")
            content_url = str(path.absolute())

        return content_url

    @classmethod
    def get_args_from_url(cls, content_url: str) -> str:
        # TODO:
        return ""



"""UniFi Protect Websockets."""

from __future__ import annotations

import asyncio
import contextlib
import logging
from collections.abc import Awaitable, Callable, Coroutine
from enum import Enum
from http import HTTPStatus
from typing import Any, Optional

import aiohttp
from aiohttp import (
    ClientError,
    ClientSession,
    ClientWebSocketResponse,
    WSMessage,
    WSMsgType,
    WSServerHandshakeError,
)
from yarl import URL

from .exceptions import NotAuthorized, NvrError

_LOGGER = logging.getLogger(__name__)
AuthCallbackType = Callable[..., Coroutine[Any, Any, Optional[dict[str, str]]]]
GetSessionCallbackType = Callable[[], Awaitable[ClientSession]]
UpdateBootstrapCallbackType = Callable[[], None]
_CLOSE_MESSAGE_TYPES = {WSMsgType.CLOSE, WSMsgType.CLOSING, WSMsgType.CLOSED}


class WebsocketState(Enum):
    CONNECTED = True
    DISCONNECTED = False


class Websocket:
    """UniFi Protect Websocket manager."""

    _running = False
    _headers: dict[str, str] | None = None
    _websocket_loop_task: asyncio.Task[None] | None = None
    _stop_task: asyncio.Task[None] | None = None
    _ws_connection: ClientWebSocketResponse | None = None

    def __init__(
        self,
        get_url: Callable[[], URL],
        auth_callback: AuthCallbackType,
        update_bootstrap: UpdateBootstrapCallbackType,
        get_session: GetSessionCallbackType,
        subscription: Callable[[WSMessage], None],
        state_callback: Callable[[WebsocketState], None],
        *,
        timeout: float = 30.0,
        backoff: int = 10,
        verify: bool = True,
        receive_timeout: float | None = None,
    ) -> None:
        """Init Websocket."""
        self.get_url = get_url
        self.timeout = timeout
        self.receive_timeout = receive_timeout
        self.backoff = backoff
        self.verify = verify
        self._get_session = get_session
        self._auth = auth_callback
        self._update_bootstrap = update_bootstrap
        self._subscription = subscription
        self._seen_non_close_message = False
        self._websocket_state = state_callback
        self._current_state: WebsocketState = WebsocketState.DISCONNECTED

    @property
    def is_connected(self) -> bool:
        """Return if the websocket is connected and has received a valid message."""
        return self._ws_connection is not None and not self._ws_connection.closed

    async def _websocket_loop(self) -> None:
        """Running loop for websocket."""
        await self.wait_closed()
        backoff = self.backoff

        while True:
            url = self.get_url()
            try:
                await self._websocket_inner_loop(url)
            except ClientError as ex:
                level = logging.ERROR if self._seen_non_close_message else logging.DEBUG
                if isinstance(ex, WSServerHandshakeError):
                    if ex.status == HTTPStatus.UNAUTHORIZED.value:
                        _LOGGER.log(
                            level, "Websocket authentication error: %s: %s", url, ex
                        )
                        await self._attempt_auth(True)
                    else:
                        _LOGGER.log(level, "Websocket handshake error: %s: %s", url, ex)
                else:
                    _LOGGER.log(level, "Websocket disconnect error: %s: %s", url, ex)
            except asyncio.TimeoutError:
                level = logging.ERROR if self._seen_non_close_message else logging.DEBUG
                _LOGGER.log(level, "Websocket timeout: %s", url)
            except Exception:
                _LOGGER.exception("Unexpected error in websocket loop")

            self._state_changed(WebsocketState.DISCONNECTED)
            if self._running is False:
                break
            _LOGGER.debug("Reconnecting websocket in %s seconds", backoff)
            await asyncio.sleep(self.backoff)

    def _state_changed(self, state: WebsocketState) -> None:
        """State changed."""
        if self._current_state is state:
            return
        self._current_state = state
        self._websocket_state(state)

    async def _websocket_inner_loop(self, url: URL) -> None:
        _LOGGER.debug("Connecting WS to %s", url)
        await self._attempt_auth(False)
        msg: WSMessage | None = None
        self._seen_non_close_message = False
        session = await self._get_session()
        # catch any and all errors for Websocket so we can clean up correctly
        try:
            self._ws_connection = await session.ws_connect(
                url,
                ssl=self.verify,
                headers=self._headers,
                timeout=aiohttp.ClientWSTimeout(ws_close=self.timeout),
            )
            while True:
                msg = await self._ws_connection.receive(self.receive_timeout)
                msg_type = msg.type
                if msg_type is WSMsgType.ERROR:
                    _LOGGER.exception("Error from Websocket: %s", msg.data)
                    break
                elif msg_type in _CLOSE_MESSAGE_TYPES:
                    _LOGGER.debug("Websocket closed: %s", msg)
                    break

                if not self._seen_non_close_message:
                    self._seen_non_close_message = True
                    self._state_changed(WebsocketState.CONNECTED)
                try:
                    self._subscription(msg)
                except Exception:
                    _LOGGER.exception("Error processing websocket message")
        finally:
            if (
                msg is not None
                and msg.type is WSMsgType.CLOSE
                # If it closes right away or lastUpdateId is in the extra
                # its an indication that we should update the bootstrap
                # since lastUpdateId is invalid
                and (
                    not self._seen_non_close_message
                    or (msg.extra and "lastUpdateId" in msg.extra)
                )
            ):
                self._update_bootstrap()
            _LOGGER.debug("Websocket disconnected: last message: %s", msg)
            if self._ws_connection is not None and not self._ws_connection.closed:
                await self._ws_connection.close()
            self._ws_connection = None

    async def _attempt_auth(self, force: bool) -> None:
        """Attempt to authenticate."""
        try:
            self._headers = await self._auth(force)
        except (NotAuthorized, NvrError) as ex:
            _LOGGER.debug("Error authenticating websocket: %s", ex)
        except Exception:
            _LOGGER.exception("Unknown error authenticating websocket")

    def start(self) -> None:
        """Start the websocket."""
        if self._running:
            return
        self._running = True
        self._websocket_loop_task = asyncio.create_task(self._websocket_loop())

    def stop(self) -> None:
        """Disconnect the websocket."""
        _LOGGER.debug("Disconnecting websocket...")
        if not self._running:
            return
        if self._websocket_loop_task:
            self._websocket_loop_task.cancel()
        self._running = False
        ws_connection = self._ws_connection
        websocket_loop_task = self._websocket_loop_task
        self._ws_connection = None
        self._websocket_loop_task = None
        self._stop_task = asyncio.create_task(
            self._stop(ws_connection, websocket_loop_task)
        )
        self._state_changed(WebsocketState.DISCONNECTED)

    async def wait_closed(self) -> None:
        """Wait for the websocket to close."""
        if self._stop_task and not self._stop_task.done():
            with contextlib.suppress(asyncio.CancelledError):
                await self._stop_task
            self._stop_task = None

    async def _stop(
        self,
        ws_connection: ClientWebSocketResponse | None,
        websocket_loop_task: asyncio.Task[None] | None,
    ) -> None:
        """Stop the websocket."""
        if ws_connection:
            await ws_connection.close()
        if websocket_loop_task:
            with contextlib.suppress(asyncio.CancelledError):
                await websocket_loop_task



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import typer

from ..api import ProtectApiClient
from ..cli import base
from ..data import Viewer

app = typer.Typer(rich_markup_mode="rich")

ARG_DEVICE_ID = typer.Argument(None, help="ID of viewer to select for subcommands")


@dataclass
class ViewerContext(base.CliContext):
    devices: dict[str, Viewer]
    device: Viewer | None = None


ALL_COMMANDS, DEVICE_COMMANDS = base.init_common_commands(app)


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context, device_id: Optional[str] = ARG_DEVICE_ID) -> None:
    """
    Viewers device CLI.

    Returns full list of Viewers without any arguments passed.
    """
    protect: ProtectApiClient = ctx.obj.protect
    context = ViewerContext(
        protect=ctx.obj.protect,
        device=None,
        devices=protect.bootstrap.viewers,
        output_format=ctx.obj.output_format,
    )
    ctx.obj = context

    if device_id is not None and device_id not in ALL_COMMANDS:
        if (device := protect.bootstrap.viewers.get(device_id)) is None:
            typer.secho("Invalid viewer ID", fg="red")
            raise typer.Exit(1)
        ctx.obj.device = device

    if not ctx.invoked_subcommand:
        if device_id in ALL_COMMANDS:
            ctx.invoke(ALL_COMMANDS[device_id], ctx)
            return

        if ctx.obj.device is not None:
            base.print_unifi_obj(ctx.obj.device, ctx.obj.output_format)
            return

        base.print_unifi_dict(ctx.obj.devices)


@app.command()
def liveview(
    ctx: typer.Context,
    liveview_id: Optional[str] = typer.Argument(None),
) -> None:
    """Returns or sets the current liveview."""
    base.require_device_id(ctx)
    obj: Viewer = ctx.obj.device

    if liveview_id is None:
        base.print_unifi_obj(obj.liveview, ctx.obj.output_format)
    else:
        protect: ProtectApiClient = ctx.obj.protect
        if (liveview_obj := protect.bootstrap.liveviews.get(liveview_id)) is None:
            typer.secho("Invalid liveview ID")
            raise typer.Exit(1)
        base.run(ctx, obj.set_liveview(liveview_obj))



from __future__ import annotations

import os

try:
    from dotenv import load_dotenv
except ImportError:
    load_dotenv = None  # type: ignore[assignment]

from .cli import app


def start() -> None:
    if load_dotenv is not None:
        env_file = os.path.join(os.getcwd(), ".env")
        if os.path.exists(env_file):
            load_dotenv(dotenv_path=env_file)
        else:
            load_dotenv()
    app()


if __name__ == "__main__":
    start()



from __future__ import annotations

from .base import (
    ProtectAdoptableDeviceModel,
    ProtectBaseObject,
    ProtectDeviceModel,
    ProtectModel,
    ProtectModelWithId,
)
from .bootstrap import Bootstrap
from .convert import create_from_unifi_dict
from .devices import (
    Bridge,
    Camera,
    CameraChannel,
    Chime,
    Doorlock,
    LCDMessage,
    Light,
    RingSetting,
    Sensor,
    Viewer,
)
from .nvr import (
    NVR,
    DoorbellMessage,
    Event,
    Liveview,
    NVRLocation,
    SmartDetectItem,
    SmartDetectTrack,
)
from .types import (
    DEFAULT,
    DEFAULT_TYPE,
    AnalyticsOption,
    AudioStyle,
    ChimeType,
    Color,
    CoordType,
    DoorbellMessageType,
    DoorbellText,
    EventCategories,
    EventType,
    FixSizeOrderedDict,
    HDRMode,
    ICRCustomValue,
    ICRLuxValue,
    IRLEDMode,
    LensType,
    LightModeEnableType,
    LightModeType,
    LockStatusType,
    ModelType,
    MountType,
    Percent,
    PermissionNode,
    ProtectWSPayloadFormat,
    PTZPosition,
    PTZPreset,
    RecordingMode,
    SensorStatusType,
    SensorType,
    SmartDetectAudioType,
    SmartDetectObjectType,
    StateType,
    StorageType,
    Version,
    VideoMode,
    WDRLevel,
)
from .user import CloudAccount, Group, Permission, User, UserLocation
from .websocket import (
    WS_HEADER_SIZE,
    WSAction,
    WSJSONPacketFrame,
    WSPacket,
    WSPacketFrameHeader,
    WSRawPacketFrame,
    WSSubscriptionMessage,
)

__all__ = [
    "DEFAULT",
    "DEFAULT_TYPE",
    "NVR",
    "WS_HEADER_SIZE",
    "AnalyticsOption",
    "AudioStyle",
    "Bootstrap",
    "Bridge",
    "Camera",
    "CameraChannel",
    "Chime",
    "ChimeType",
    "CloudAccount",
    "Color",
    "CoordType",
    "DoorbellMessage",
    "DoorbellMessageType",
    "DoorbellText",
    "Doorlock",
    "Event",
    "EventCategories",
    "EventType",
    "FixSizeOrderedDict",
    "Group",
    "HDRMode",
    "ICRCustomValue",
    "ICRLuxValue",
    "IRLEDMode",
    "LCDMessage",
    "LensType",
    "Light",
    "LightModeEnableType",
    "LightModeType",
    "Liveview",
    "LockStatusType",
    "ModelType",
    "MountType",
    "NVRLocation",
    "PTZPosition",
    "PTZPreset",
    "Percent",
    "Permission",
    "PermissionNode",
    "ProtectAdoptableDeviceModel",
    "ProtectBaseObject",
    "ProtectDeviceModel",
    "ProtectModel",
    "ProtectModelWithId",
    "ProtectWSPayloadFormat",
    "RecordingMode",
    "RingSetting",
    "Sensor",
    "SensorStatusType",
    "SensorType",
    "SmartDetectAudioType",
    "SmartDetectItem",
    "SmartDetectObjectType",
    "SmartDetectTrack",
    "StateType",
    "StorageType",
    "User",
    "UserLocation",
    "Version",
    "VideoMode",
    "Viewer",
    "WDRLevel",
    "WSAction",
    "WSJSONPacketFrame",
    "WSPacket",
    "WSPacketFrameHeader",
    "WSRawPacketFrame",
    "WSSubscriptionMessage",
    "create_from_unifi_dict",
]



from __future__ import annotations

import asyncio
import base64
import logging
import sys
from pathlib import Path
from typing import Optional, cast

import orjson
import typer
from rich.progress import track

from uiprotect.api import ProtectApiClient

from ..data import Version, WSPacket
from ..test_util import SampleDataGenerator
from ..utils import RELEASE_CACHE, get_local_timezone, run_async
from ..utils import profile_ws as profile_ws_job
from .base import CliContext, OutputFormatEnum
from .cameras import app as camera_app
from .chimes import app as chime_app
from .doorlocks import app as doorlock_app
from .events import app as event_app
from .lights import app as light_app
from .liveviews import app as liveview_app
from .nvr import app as nvr_app
from .sensors import app as sensor_app
from .viewers import app as viewer_app

try:
    from .backup import app as backup_app
except ImportError:
    backup_app = None  # type: ignore[assignment]

_LOGGER = logging.getLogger("uiprotect")

try:
    from IPython import embed
    from termcolor import colored
    from traitlets.config import get_config
except ImportError:
    embed = termcolor = get_config = None  # type: ignore[assignment]

OPTION_USERNAME = typer.Option(
    ...,
    "--username",
    "-U",
    help="UniFi Protect username",
    prompt=True,
    envvar="UFP_USERNAME",
)
OPTION_PASSWORD = typer.Option(
    ...,
    "--password",
    "-P",
    help="UniFi Protect password",
    prompt=True,
    hide_input=True,
    envvar="UFP_PASSWORD",
)
OPTION_ADDRESS = typer.Option(
    ...,
    "--address",
    "-a",
    prompt=True,
    help="UniFi Protect IP address or hostname",
    envvar="UFP_ADDRESS",
)
OPTION_PORT = typer.Option(
    443,
    "--port",
    "-p",
    help="UniFi Protect Port",
    envvar="UFP_PORT",
)
OPTION_SECONDS = typer.Option(15, "--seconds", "-s", help="Seconds to pull events")
OPTION_VERIFY = typer.Option(
    True,
    "--no-verify",
    help="Verify SSL",
    envvar="UFP_SSL_VERIFY",
)
OPTION_ANON = typer.Option(True, "--actual", help="Do not anonymize test data")
OPTION_ZIP = typer.Option(False, "--zip", help="Zip up data after generate")
OPTION_WAIT = typer.Option(
    30,
    "--wait",
    "-w",
    help="Time to wait for Websocket messages",
)
OPTION_OUTPUT = typer.Option(
    None,
    "--output",
    "-o",
    help="Output folder, defaults to `tests` folder one level above this file",
    envvar="UFP_SAMPLE_DIR",
)
OPTION_OUT_FORMAT = typer.Option(
    OutputFormatEnum.PLAIN,
    "--output-format",
    help="Preferred output format. Not all commands support both JSON and plain and may still output in one or the other.",
)
OPTION_WS_FILE = typer.Option(
    None,
    "--file",
    "-f",
    help="Path or raw binary Websocket message",
)
OPTION_UNADOPTED = typer.Option(
    False,
    "-u",
    "--include-unadopted",
    help="Include devices not adopted by this NVR.",
)
ARG_WS_DATA = typer.Argument(None, help="base64 encoded Websocket message")

SLEEP_INTERVAL = 2


app = typer.Typer(rich_markup_mode="rich")
app.add_typer(nvr_app, name="nvr")
app.add_typer(event_app, name="events")
app.add_typer(liveview_app, name="liveviews")
app.add_typer(camera_app, name="cameras")
app.add_typer(chime_app, name="chimes")
app.add_typer(doorlock_app, name="doorlocks")
app.add_typer(light_app, name="lights")
app.add_typer(sensor_app, name="sensors")
app.add_typer(viewer_app, name="viewers")

if backup_app is not None:
    app.add_typer(backup_app, name="backup")


@app.callback()
def main(
    ctx: typer.Context,
    username: str = OPTION_USERNAME,
    password: str = OPTION_PASSWORD,
    address: str = OPTION_ADDRESS,
    port: int = OPTION_PORT,
    verify: bool = OPTION_VERIFY,
    output_format: OutputFormatEnum = OPTION_OUT_FORMAT,
    include_unadopted: bool = OPTION_UNADOPTED,
) -> None:
    """UniFi Protect CLI"""
    # preload the timezone before any async code runs
    get_local_timezone()

    protect = ProtectApiClient(
        address,
        port,
        username,
        password,
        verify_ssl=verify,
        ignore_unadopted=not include_unadopted,
    )

    async def update() -> None:
        protect._bootstrap = await protect.get_bootstrap()
        await protect.close_session()

    run_async(update())
    ctx.obj = CliContext(protect=protect, output_format=output_format)


def _setup_logger(level: int = logging.DEBUG, show_level: bool = False) -> None:
    console_handler = logging.StreamHandler()
    console_handler.setLevel(level)
    if show_level:
        formatter = logging.Formatter("%(levelname)s: %(message)s")
        console_handler.setFormatter(formatter)
    _LOGGER.setLevel(logging.DEBUG)
    _LOGGER.addHandler(console_handler)


async def _progress_bar(wait_time: int, label: str) -> None:
    for i in track(range(wait_time // SLEEP_INTERVAL), description=label):
        if i > 0:
            await asyncio.sleep(SLEEP_INTERVAL)


@app.command()
def shell(ctx: typer.Context) -> None:
    """
    Opens iPython shell with Protect client initialized.

    Requires the `shell` extra to also be installed.
    """
    if embed is None or colored is None:
        typer.echo("ipython and termcolor required for shell subcommand")
        sys.exit(1)

    # locals passed to shell
    protect = cast(
        ProtectApiClient,
        ctx.obj.protect,
    )
    _setup_logger(show_level=True)

    async def wait_forever() -> None:
        await protect.update()
        protect.subscribe_websocket(lambda _: None)
        while True:
            await asyncio.sleep(10)
            await protect.update()

    c = get_config()
    c.InteractiveShellEmbed.colors = "Linux"
    embed(  # type: ignore[no-untyped-call]
        header=colored("protect = ProtectApiClient(*args)", "green"),
        config=c,
        using="asyncio",
    )


@app.command()
def generate_sample_data(
    ctx: typer.Context,
    anonymize: bool = OPTION_ANON,
    wait_time: int = OPTION_WAIT,
    output_folder: Optional[Path] = OPTION_OUTPUT,
    do_zip: bool = OPTION_ZIP,
) -> None:
    """Generates sample data for UniFi Protect instance."""
    protect = cast(ProtectApiClient, ctx.obj.protect)

    if output_folder is None:
        tests_folder = Path(__file__).parent.parent / "tests"

        if not tests_folder.exists():
            typer.secho("Output folder required when not in dev-mode", fg="red")
            sys.exit(1)
        output_folder = (tests_folder / "sample_data").absolute()

    def log(msg: str) -> None:
        typer.echo(msg)

    def log_warning(msg: str) -> None:
        typer.secho(msg, fg="yellow")

    SampleDataGenerator(
        protect,
        output_folder,
        anonymize,
        wait_time,
        log=log,
        log_warning=log_warning,
        ws_progress=_progress_bar,
        do_zip=do_zip,
    ).generate()


@app.command()
def profile_ws(
    ctx: typer.Context,
    wait_time: int = OPTION_WAIT,
    output_path: Optional[Path] = OPTION_OUTPUT,
) -> None:
    """Profiles Websocket messages for UniFi Protect instance."""
    protect = cast(ProtectApiClient, ctx.obj.protect)

    async def callback() -> None:
        await protect.update()
        unsub = protect.subscribe_websocket(lambda _: None)
        await profile_ws_job(
            protect,
            wait_time,
            output_path=output_path,
            ws_progress=_progress_bar,
        )
        unsub()
        await protect.async_disconnect_ws()
        await protect.close_session()

    _setup_logger()

    run_async(callback())


@app.command()
def decode_ws_msg(
    ws_file: typer.FileBinaryRead = OPTION_WS_FILE,
    ws_data: Optional[str] = ARG_WS_DATA,
) -> None:
    """Decodes a base64 encoded UniFi Protect Websocket binary message."""
    if ws_file is None and ws_data is None:  # type: ignore[unreachable]
        typer.secho("Websocket data required", fg="red")  # type: ignore[unreachable]
        sys.exit(1)

    ws_data_raw = b""
    if ws_file is not None:
        ws_data_raw = ws_file.read()
    elif ws_data is not None:  # type: ignore[unreachable]
        ws_data_raw = base64.b64decode(ws_data.encode("utf8"))

    packet = WSPacket(ws_data_raw)
    response = {"action": packet.action_frame.data, "data": packet.data_frame.data}

    typer.echo(orjson.dumps(response).decode("utf-8"))


@app.command()
def release_versions(ctx: typer.Context) -> None:
    """Updates the release version cache on disk."""
    protect = cast(ProtectApiClient, ctx.obj.protect)

    async def callback() -> set[Version]:
        versions = await protect.get_release_versions()
        await protect.close_session()
        return versions

    _setup_logger()

    versions = run_async(callback())
    output = orjson.dumps(sorted([str(v) for v in versions]))

    Path(RELEASE_CACHE).write_bytes(output)
    typer.echo(output.decode("utf-8"))



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import typer

from ..api import ProtectApiClient
from ..cli import base
from ..data import Liveview

app = typer.Typer(rich_markup_mode="rich")

ARG_DEVICE_ID = typer.Argument(None, help="ID of liveview to select for subcommands")
ALL_COMMANDS = {"list-ids": app.command(name="list-ids")(base.list_ids)}

app.command(name="protect-url")(base.protect_url)


@dataclass
class LiveviewContext(base.CliContext):
    devices: dict[str, Liveview]
    device: Liveview | None = None


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context, device_id: Optional[str] = ARG_DEVICE_ID) -> None:
    """
    Liveviews CLI.

    Returns full list of Liveviews without any arguments passed.
    """
    protect: ProtectApiClient = ctx.obj.protect
    context = LiveviewContext(
        protect=ctx.obj.protect,
        device=None,
        devices=protect.bootstrap.liveviews,
        output_format=ctx.obj.output_format,
    )
    ctx.obj = context

    if device_id is not None and device_id not in ALL_COMMANDS:
        if (device := protect.bootstrap.liveviews.get(device_id)) is None:
            typer.secho("Invalid liveview ID", fg="red")
            raise typer.Exit(1)
        ctx.obj.device = device

    if not ctx.invoked_subcommand:
        if device_id in ALL_COMMANDS:
            ctx.invoke(ALL_COMMANDS[device_id], ctx)
            return

        if ctx.obj.device is not None:
            base.print_unifi_obj(ctx.obj.device, ctx.obj.output_format)
            return

        base.print_unifi_dict(ctx.obj.devices)


@app.command()
def owner(ctx: typer.Context) -> None:
    """Gets the owner for the liveview."""
    base.require_device_id(ctx)
    obj: Liveview = ctx.obj.device
    base.print_unifi_obj(obj.owner, ctx.obj.output_format)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import typer

from ..api import ProtectApiClient
from ..cli import base
from ..data import Chime

app = typer.Typer(rich_markup_mode="rich")

ARG_DEVICE_ID = typer.Argument(None, help="ID of chime to select for subcommands")
ARG_REPEAT = typer.Argument(..., help="Repeat times count", min=1, max=6)
ARG_VOLUME = typer.Argument(..., help="Volume", min=1, max=100)


@dataclass
class ChimeContext(base.CliContext):
    devices: dict[str, Chime]
    device: Chime | None = None


ALL_COMMANDS, DEVICE_COMMANDS = base.init_common_commands(app)


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context, device_id: Optional[str] = ARG_DEVICE_ID) -> None:
    """
    Chime device CLI.

    Returns full list of Chimes without any arguments passed.
    """
    protect: ProtectApiClient = ctx.obj.protect
    context = ChimeContext(
        protect=ctx.obj.protect,
        device=None,
        devices=protect.bootstrap.chimes,
        output_format=ctx.obj.output_format,
    )
    ctx.obj = context

    if device_id is not None and device_id not in ALL_COMMANDS:
        if (device := protect.bootstrap.chimes.get(device_id)) is None:
            typer.secho("Invalid chime ID", fg="red")
            raise typer.Exit(1)
        ctx.obj.device = device

    if not ctx.invoked_subcommand:
        if device_id in ALL_COMMANDS:
            ctx.invoke(ALL_COMMANDS[device_id], ctx)
            return

        if ctx.obj.device is not None:
            base.print_unifi_obj(ctx.obj.device, ctx.obj.output_format)
            return

        base.print_unifi_dict(ctx.obj.devices)


@app.command()
def cameras(
    ctx: typer.Context,
    camera_ids: list[str] = typer.Argument(
        None,
        help="Set to [] to empty list of cameras",
    ),
    add: bool = typer.Option(False, "-a", "--add", help="Add cameras instead of set"),
    remove: bool = typer.Option(
        False,
        "-r",
        "--remove",
        help="Remove cameras instead of set",
    ),
) -> None:
    """Returns or sets paired doorbells for the chime."""
    base.require_device_id(ctx)
    obj: Chime = ctx.obj.device

    if add and remove:
        typer.secho("Add and remove are mutally exclusive", fg="red")
        raise typer.Exit(1)

    if len(camera_ids) == 0:
        base.print_unifi_list(obj.cameras)
        return

    protect: ProtectApiClient = ctx.obj.protect

    if len(camera_ids) == 1 and camera_ids[0] == "[]":
        camera_ids = []

    for camera_id in camera_ids:
        if (camera := protect.bootstrap.cameras.get(camera_id)) is None:
            typer.secho(f"Invalid camera ID: {camera_id}", fg="red")
            raise typer.Exit(1)

        if not camera.feature_flags.is_doorbell:
            typer.secho(f"Camera is not a doorbell: {camera_id}", fg="red")
            raise typer.Exit(1)

    if add:
        camera_ids = list(set(obj.camera_ids) | set(camera_ids))
    elif remove:
        camera_ids = list(set(obj.camera_ids) - set(camera_ids))

    data_before_changes = obj.dict_with_excludes()
    obj.camera_ids = camera_ids
    base.run(ctx, obj.save_device(data_before_changes))


@app.command()
def set_volume(
    ctx: typer.Context,
    value: int = ARG_VOLUME,
    camera_id: Optional[str] = typer.Option(
        None,
        "-c",
        "--camera",
        help="Camera ID to apply volume to",
    ),
) -> None:
    """Set volume level for chime rings."""
    base.require_device_id(ctx)
    obj: Chime = ctx.obj.device
    if camera_id is None:
        base.run(ctx, obj.set_volume(value))
    else:
        protect: ProtectApiClient = ctx.obj.protect
        camera = protect.bootstrap.cameras.get(camera_id)
        if camera is None:
            typer.secho(f"Invalid camera ID: {camera_id}", fg="red")
            raise typer.Exit(1)
        base.run(ctx, obj.set_volume_for_camera(camera, value))


@app.command()
def play(
    ctx: typer.Context,
    volume: Optional[int] = typer.Option(None, "-v", "--volume", min=1, max=100),
    repeat_times: Optional[int] = typer.Option(None, "-r", "--repeat", min=1, max=6),
) -> None:
    """Plays chime tone."""
    base.require_device_id(ctx)
    obj: Chime = ctx.obj.device
    base.run(ctx, obj.play(volume=volume, repeat_times=repeat_times))


@app.command()
def play_buzzer(ctx: typer.Context) -> None:
    """Plays chime buzzer."""
    base.require_device_id(ctx)
    obj: Chime = ctx.obj.device
    base.run(ctx, obj.play_buzzer())


@app.command()
def set_repeat_times(
    ctx: typer.Context,
    value: int = ARG_REPEAT,
    camera_id: Optional[str] = typer.Option(
        None,
        "-c",
        "--camera",
        help="Camera ID to apply repeat times to",
    ),
) -> None:
    """Set number of times for a chime to repeat when doorbell is rang."""
    base.require_device_id(ctx)
    obj: Chime = ctx.obj.device
    if camera_id is None:
        base.run(ctx, obj.set_repeat_times(value))
    else:
        protect: ProtectApiClient = ctx.obj.protect
        camera = protect.bootstrap.cameras.get(camera_id)
        if camera is None:
            typer.secho(f"Invalid camera ID: {camera_id}", fg="red")
            raise typer.Exit(1)
        base.run(ctx, obj.set_repeat_times_for_camera(camera, value))
